# -*- coding: utf-8 -*-
"""Specifies the version of the Deep Time-Series package."""

__title__ = "autotime"
__author__ = "Kevin R. Chen"
__description__ = "Automated ML-based predictive analytics framework for time-series data."
__url__ = "https://github.com/Kevin-Chen0/auto-time-series.git"
__version__ = "0.1.3.1"
__license__ = "Apache License 2.0"
__copyright__ = "2020 Kevin R. Chen"
